from ._bridges import SpectralBridges  # noqa: D104
from ._defs import ExpQuantileTransform, ngap_scorer, silhouette_scorer

__all__ = [
    "ExpQuantileTransform",
    "SpectralBridges",
    "ngap_scorer",
    "silhouette_scorer",
]



from __future__ import annotations

from numbers import Real
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

import numpy as np
from sklearn.metrics import silhouette_score  # type: ignore
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore

if TYPE_CHECKING:
    from ._bridges import SpectralBridges


# Protocols
@runtime_checkable
class AffinityTransform(Protocol):
    """Protocol for affinity transforms.

    Use this protocol to define custom affinity transforms.
    """

    def __call__(self, x: np.ndarray) -> np.ndarray: ...


# Transformations
class ExpQuantileTransform(AffinityTransform):
    """Exponential quantile transform.

    Attributes:
        alphas (tuple[float, float]): Quantiles for affinity matrix computation.
        mult_factor (int | float): Scaling parameter for affinity matrix computation.
    """

    q1: float
    q2: float
    mult_factor: float

    @validate_params(
        {
            "q1": [Interval(Real, 0, 1, closed="left")],
            "q2": [Interval(Real, 0, 1, closed="left")],
            "mult_factor": [Interval(Real, 0, None, closed="left")],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(self, q1: float, q2: float, mult_factor: float):
        """Initialize the Exponential quantile transform.

        Args:
            q1 (float): First quantile for affinity matrix computation.
            q2 (float): Second quantile for affinity matrix computation.
            mult_factor (float): Scaling parameter for affinity matrix
                computation.
        """
        if not (0 <= q1 < q2 <= 1):
            raise ValueError("q1 and q2 must be between 0 and 1 and q1 < q2")

        self.q1 = q1
        self.q2 = q2
        self.mult_factor = mult_factor

    def __call__(self, x: np.ndarray) -> np.ndarray:
        q1, q2 = np.quantile(x, (self.q1, self.q2))
        gamma = np.log(self.mult_factor) / (q2 - q1)
        return np.exp(gamma * (x - x.max()))


# Scorers
def silhouette_scorer(estimator: SpectralBridges, X: np.typing.ArrayLike) -> float:
    """Silhouette scorer for SpectralBridges.

    Args:
        estimator (SpectralBridges): The estimator to score.
        X (np.typing.ArrayLike): The data to score.

    Returns:
        float: The silhouette score.
    """
    return silhouette_score(X, estimator.predict(X))


def ngap_scorer(estimator: SpectralBridges, *_args: Any, **_kwargs: Any) -> float:
    """NGAP scorer for SpectralBridges.

    Args:
        estimator (SpectralBridges): The estimator to score.
        X (np.typing.ArrayLike): The data to score.

    Returns:
        float: The NGAP score.
    """
    return estimator.ngap_ if estimator.ngap_ is not None else -np.inf



from numbers import Integral, Real
from typing import Any, Final, Self, cast

import faiss  # type: ignore
import numpy as np
from scipy.linalg import eigh  # type: ignore
from scipy.linalg.blas import dgemm  # type: ignore
from sklearn.base import BaseEstimator, ClusterMixin  # type: ignore
from sklearn.utils._param_validation import Interval, validate_params  # type: ignore
from sklearn.utils.validation import check_is_fitted  # type: ignore

from ._defs import AffinityTransform, ExpQuantileTransform
from ._kmeans import KMeans

# Constants
DEFAULT_AFFINITY_TRANSFORM: Final = ExpQuantileTransform(0.1, 0.9, 1e4)


class SpectralBridges(BaseEstimator, ClusterMixin):
    """Spectral Bridges clustering algorithm.

    Attributes:
        random_state (int | None): Determines random number generation for centroid
            initialization.
        n_clusters (int): The number of clusters to form.
        n_nodes (int | None): Number of nodes or initial clusters.
        p (int | float): Power of the alpha_i.
        n_iter (int): Number of iterations to run the k-means algorithm.
        n_local_trials (int | None): Number of seeding trials for centroids
            initialization.
        random_state (int | None): Determines random number generation for centroid
            initialization.
        tol (float): Tolerance for the normalized eigengap.
        affinity_transform (AffinityTransform): Affinity transform to apply to the
            affinity matrix.
        cluster_centers_ (list[np.ndarray] | None): Coordinates of cluster
            centers.
        labels_ (np.ndarray): Labels of each cluster (index) in X.
        ngap_ (float): The normalized eigengap.
    """

    n_clusters: int
    n_nodes: int | None
    p: int | float
    n_iter: int
    n_local_trials: int | None
    random_state: int | None
    tol: float
    cluster_centers_: list[np.ndarray] | None
    labels_: np.ndarray | None
    ngap_: float | None
    affinity_transform: AffinityTransform

    @validate_params(
        {
            "X": ["array-like"],
            "n_clusters": [Interval(Integral, 2, None, closed="left")],
            "n_nodes": [Interval(Integral, 2, None, closed="left"), None],
            "p": [Interval(Real, 0, None, closed="left")],
            "n_iter": [Interval(Integral, 1, None, closed="left")],
            "n_local_trials": [Interval(Integral, 1, None, closed="left"), None],
            "random_state": ["random_state"],
            "tol": [Interval(Real, 0, None, closed="left")],
            "affinity_transform": [AffinityTransform],
        },
        prefer_skip_nested_validation=True,
    )
    def __init__(
        self,
        n_clusters: int,
        n_nodes: int | None = None,
        *,
        p: int | float = 2,
        n_iter: int = 20,
        n_local_trials: int | None = None,
        random_state: int | None = None,
        tol: float = 1e-8,
        affinity_transform: AffinityTransform = DEFAULT_AFFINITY_TRANSFORM,
    ):
        """Initialize the Spectral Bridges model.

        Args:
            n_clusters (int): The number of clusters to form.
            n_nodes  (int | None): Number of nodes or initial clusters.
            p (int | float, optional): Power of the alpha_i. Defaults to 2.
            n_iter (int, optional): Number of iterations to run the k-means
                algorithm. Defaults to 20.
            n_local_trials (int | None, optional): Number of seeding trials for
                centroids initialization. Defaults to None.
            random_state (int | None, optional): Determines random number
                generation for centroid initialization. Defaults to None.
            tol (float, optional): Tolerance for the normalized eigengap.
                Defaults to 1e-8.
            affinity_transform (AffinityTransform, optional): Affinity transform
                to apply to the affinity matrix. Defaults to DEFAULT_AFFINITY_TRANSFORM.
        """
        self.n_clusters = n_clusters
        self.n_nodes = n_nodes
        self.p = p
        self.n_iter = n_iter
        self.n_local_trials = n_local_trials
        self.random_state = random_state
        self.tol = tol
        self.affinity_transform = affinity_transform
        self.cluster_centers_ = None
        self.ngap_ = None

        if self.n_nodes is not None and self.n_nodes <= self.n_clusters:
            raise ValueError(
                f"n_nodes must be greater than n_clusters, got {self.n_nodes} <= "
                f"{self.n_clusters}"
            )

    @validate_params(
        {
            "X": ["array-like"],
            "y": [None],
        },
        prefer_skip_nested_validation=True,
    )
    def fit(self, X: np.typing.ArrayLike, y: None = None) -> Self:  # noqa: ARG002
        """Fit the Spectral Bridges model on the input data X.

        Args:
            X (np.typing.ArrayLike): Input data to cluster.
            y (None, optional): Placeholder for y.

        Raises:
            ValueError: If the number of samples is less than the number of clusters.
            ValueError: If n_nodes is not provided.

        Returns:
            Self: The fitted model.
        """
        X = np.asarray(X)  # type: ignore

        if X.shape[0] < self.n_clusters:
            raise ValueError(
                f"n_samples={X.shape[0]} must be >= n_clusters={self.n_clusters}."
            )
        if self.n_nodes is None:
            raise ValueError("n_nodes must be provided")

        kmeans = KMeans(
            self.n_nodes,
            self.n_iter,
            self.n_local_trials,
            self.random_state,
        )
        kmeans.fit(X)
        centers = cast(np.ndarray, kmeans.cluster_centers_)

        affinity = np.empty((self.n_nodes, self.n_nodes), dtype=np.float64)

        X_centered = [
            np.array(
                X[kmeans.labels_ == i] - cast(np.ndarray, kmeans.cluster_centers_)[i],
                dtype=np.float64,
                order="F",
            )
            for i in range(self.n_nodes)
        ]

        counts = np.array([X_centered[i].shape[0] for i in range(self.n_nodes)])
        counts = counts[None, :] + counts[:, None]

        for i in range(self.n_nodes):
            segments = np.asfortranarray(centers - centers[i])
            dists = np.einsum("ij,ij->i", segments, segments)
            dists[i] = 1

            projs = cast(np.ndarray, dgemm(1.0, X_centered[i], segments, trans_b=True))
            np.clip(projs / dists, 0, None, out=projs)
            projs = np.power(projs, self.p)

            affinity[i] = projs.sum(axis=0)

        affinity = np.power((affinity + affinity.T) / counts, 1 / self.p)

        affinity = self.affinity_transform(affinity)

        d = np.power(affinity.mean(axis=1), -0.5)
        L = -(d[:, None] * affinity * d[None, :])
        np.fill_diagonal(L, self.n_nodes + self.tol)

        eigvals, eigvecs = cast(
            tuple[np.ndarray, np.ndarray],
            eigh(
                L,
                subset_by_index=[0, self.n_clusters],
            ),
        )
        eigvecs = eigvecs[:, :-1]
        eigvecs /= np.linalg.norm(eigvecs, axis=1)[:, None]
        self.labels_ = (
            KMeans(self.n_clusters, self.n_iter, self.n_local_trials, self.random_state)
            .fit(eigvecs)
            .labels_
        )

        self.ngap_ = (eigvals[-1] - eigvals[-2]) / eigvals[-2]
        self.cluster_centers_ = [
            centers[self.labels_ == i] for i in range(self.n_clusters)
        ]

        return self

    @validate_params(
        {
            "X": ["array-like"],
        },
        prefer_skip_nested_validation=True,
    )
    def predict(self, X: np.typing.ArrayLike) -> np.ndarray:
        """Predict the nearest cluster index for each input data point x.

        Args:
            X (np.typing.ArrayLike): The input data.

        Raises:
            ValueError: If `X` contains inf or NaN values.
            ValueError: If `X` contains inf or NaN values.
            ValueError: If `self.cluster_centers_` is not set.

        Returns:
            np.ndarray The predicted cluster indices.
        """
        check_is_fitted(self)

        X_f32 = np.asarray(X).astype(np.float32)

        if np.isinf(X_f32).any():
            raise ValueError("X must not contain inf values")
        if np.isnan(X_f32).any():
            raise ValueError("X must not contain NaN values")
        if self.cluster_centers_ is None:
            raise ValueError("SpectralBridges must be fitted first")

        cluster_centers = np.vstack(self.cluster_centers_)
        cluster_cutoffs = np.cumsum(
            [cluster.shape[0] for cluster in self.cluster_centers_]
        )

        index = faiss.IndexFlatL2(X_f32.shape[1])
        index.add(cluster_centers)  # type: ignore
        winners = cast(np.ndarray, index.search(X_f32, 1)[1]).ravel()  # type: ignore

        return np.searchsorted(cluster_cutoffs, winners, side="right")

    @validate_params(
        {
            "X": ["array-like"],
            "y": [None],
        },
        prefer_skip_nested_validation=True,
    )
    def fit_predict(
        self,
        X: np.typing.ArrayLike,
        y: None = None,  # noqa: ARG002
        **_kwargs: Any,
    ) -> np.ndarray:
        """Fit the model and predict the nearest cluster index for each input point.

        Args:
            X (np.typing.ArrayLike): The input data.
            y (None, optional): Placeholder for y.

        Returns:
            np.ndarray: The predicted cluster indices.
        """
        return self.fit(X).predict(X)



from typing import Self, cast

import faiss  # type: ignore
import numpy as np
from scipy.linalg.blas import sgemm  # type: ignore


class KMeans:
    """K-means clustering using FAISS.

    Attributes:
        n_clusters (int): The number of clusters to form.
        n_iter (int): The number of iterations to run the k-means
            algorithm.
        n_local_trials  (int | None): The number of seeding trials for
            centroids initialization.
        random_state (int | None) Determines random number generation for
            centroid initialization.
        cluster_centers_ (np.ndarray | None): Coordinates of cluster centers.
        labels_ (np.ndarray | None): Labels of each point (index) in X.
    """

    n_clusters: int
    n_iter: int
    n_local_trials: int | None
    random_state: int | None
    cluster_centers_: np.ndarray | None
    labels_: np.ndarray | None

    def __init__(
        self,
        n_clusters: int,
        n_iter: int,
        n_local_trials: int | None,
        random_state: int | None,
    ):
        """Initializes the KMeans class.

        Args:
            n_clusters (int): The number of clusters to form.
            n_iter (int): The number of iterations to run the k-means
                algorithm.
            n_local_trials  (int | None): The number of seeding trials for
                centroids initialization.
            random_state (int | None) Determines random number generation for
                centroid initialization.
        """
        self.n_clusters = n_clusters
        self.n_iter = n_iter
        self.n_local_trials = n_local_trials
        self.random_state = random_state
        self.cluster_centers_ = None
        self.labels_ = None

    @staticmethod
    def _dists(X: np.ndarray, y: np.ndarray, XX: np.ndarray) -> np.ndarray:
        """Computes the pairwise distances between a fixed data matrix and some points.

        Args:
            X (np.ndarray): The fixed data matrix.
            y (np.ndarray): The non fixed points.
            XX (np.ndarray): The fixed matrix squared norm.

        Returns:
            np.ndarray: The computed pairwise distances.
        """
        yy = np.einsum("ij,ij->i", y, y)
        dists = XX - cast(np.ndarray, sgemm(2.0, X, y, trans_b=True)) + yy
        np.clip(dists, 0, None, out=dists)

        return dists

    def _init_centroids(self, X: np.ndarray) -> np.ndarray:
        """Initializes the centroids in a K-means++ fashion.

        Args:
            X (np.ndarray): The fixed data matrix.

        Returns:
            np.ndarray: The initialized centroids.
        """
        rng = np.random.default_rng(self.random_state)

        centroids = np.empty((self.n_clusters, X.shape[1]), dtype=X.dtype)
        centroids[0] = X[rng.integers(X.shape[0])]

        XX = np.einsum("ij,ij->i", X, X)[:, None]

        dists = self._dists(X, centroids[:1], XX).ravel()
        inertia = dists.sum()

        if self.n_local_trials is None:
            self.n_local_trials = 2 + int(np.log(self.n_clusters))

        for i in range(1, self.n_clusters):
            candidate_ids = rng.choice(
                X.shape[0], size=self.n_local_trials, p=dists / inertia
            )
            candidates = np.asfortranarray(X[candidate_ids])

            current_candidates_dists = self._dists(X, candidates, XX)
            candidates_dists = np.minimum(current_candidates_dists, dists[:, None])

            inertias = candidates_dists.sum(axis=0)
            best_inertia = inertias.argmin()
            best_candidate = candidate_ids[best_inertia]
            dists = candidates_dists[:, best_inertia]
            inertia = inertias[best_inertia]

            centroids[i] = X[best_candidate]

        return centroids

    @staticmethod
    def _validate_X(X: np.ndarray) -> np.ndarray:
        """Validates and converts the data matrix.

        Args:
            X (NDArray[np.float32  |  np.float64]): The fixed data matrix.

        Raises:
            ValueError: If `X``contains inf values.
            ValueError: If `X``contains NaN values.

        Returns:
            np.ndarray: The validated and converted data matrix.
        """
        X_f32 = X.astype(np.float32, order="F")

        if np.isinf(X_f32).any():
            raise ValueError("X must not contain inf values")
        if np.isnan(X_f32).any():
            raise ValueError("X must not contain NaN values")

        return X_f32

    def fit(self, X: np.ndarray) -> Self:
        """Run k-means clustering on the input data X.

        Args:
            X (np.ndarray): Input data matrix to cluster.

        Returns:
            Self: The fitted model.
        """
        X_f32 = self._validate_X(X)

        index = faiss.IndexFlatL2(X_f32.shape[1])
        kmeans = faiss.Clustering(X_f32.shape[1], self.n_clusters)

        init_centroids = self._init_centroids(X_f32)

        kmeans.centroids.resize(init_centroids.size)
        faiss.copy_array_to_vector(init_centroids.ravel(), kmeans.centroids)  # type: ignore
        kmeans.niter = self.n_iter
        kmeans.min_points_per_centroid = 0
        kmeans.max_points_per_centroid = -1
        kmeans.train(X_f32, index)  # type: ignore

        self.cluster_centers_ = cast(
            np.ndarray,
            faiss.vector_to_array(kmeans.centroids).reshape(  # type: ignore
                self.n_clusters, X_f32.shape[1]
            ),
        )
        self.labels_ = cast(np.ndarray, index.search(X_f32, 1)[1].ravel())  # type: ignore

        return self
